'use strict';

var _ = require('lodash');
var Qs = require('qs');
var co = require('co');
var parseBody = require('co-body');
var KindaObject = require('kinda-object');
var KindaEventManager = require('kinda-event-manager');
var util = require('kinda-util').create();

// TODO: authorization verifier, authorize handler, custom methods and event listeners should inherit from super classes

var RegisteredCollection = KindaObject.extend('RegisteredCollection', function () {
  this.include(KindaEventManager);

  this.creator = function (name) {
    var _this = this;

    var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];

    this.name = name;
    this.slug = _.kebabCase(name);
    this.verifyAuthorizationHandler = options.verifyAuthorizationHandler;
    this.authorizeHandler = options.authorizeHandler;
    this.collectionMethods = options.collectionMethods || {};
    this.itemMethods = options.itemMethods || {};
    _.forOwn(options.eventListeners, function (fn, event) {
      _this.on(event, fn);
    });
  };
});

var KindaRepositoryServer = KindaObject.extend('KindaRepositoryServer', function () {
  this.include(KindaEventManager);

  // Options:
  //   signInWithCredentialsHandler:
  //     async function(credentials) {
  //       if (!credentials) return;
  //       if (credentials.username !== 'user@domain.com') return;
  //       if (credentials.password !== 'password') return;
  //       return 'secret-token';
  //     }
  //   signInWithAuthorizationHandler:
  //     async function(authorization) {
  //       return authorization === 'secret-token';
  //     },
  //   signOutHandler:
  //     async function(authorization) {
  //       // delete authorization token...
  //     },
  //   verifyAuthorizationHandler:
  //     async function(request) {
  //       return request.authorization === 'secret-token';
  //     }
  //   authorizeHandler:
  //     async function(request) {
  //       return request.verifyAuthorizationResult || request.method === 'GET';
  //     }
  this.creator = function () {
    var _this2 = this;

    var options = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

    if (!options.repository) throw new Error('repository is missing');
    this.repository = options.repository;
    this.remoteRepository = options.remoteRepository || options.repository;
    _.assign(this, _.pick(options, ['signInWithCredentialsHandler', 'signInWithAuthorizationHandler', 'signOutHandler', 'verifyAuthorizationHandler', 'authorizeHandler']));
    this.registeredCollections = {};
    var collectionOptions = options.collections || {};
    _.forOwn(this.remoteRepository.collectionClasses, function (klass, name) {
      if (!_this2.repository.collectionClasses[name]) {
        throw new Error('collection \'' + name + '\' is undefined in the server repository');
      }
      _this2.registerCollection(name, collectionOptions[name]);
    });
  };

  this.use = function (plugin) {
    plugin.plug(this);
  };

  this.registerCollection = function (name, options) {
    var collection = RegisteredCollection.create(name, options);
    this.registeredCollections[collection.slug] = collection;
  };

  this.handleRequest = function callee$1$0(ctx, path, next) {
    var query, slug, index;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          query = Qs.parse(ctx.querystring);

          query = util.decodeValue(query);
          ctx.options = query;
          ctx.authorization = this.authorizationUnserializer({ query: query });

          slug = path;

          if (_.startsWith(slug, '/')) slug = slug.slice(1);
          index = slug.indexOf('/');

          if (index !== -1) {
            path = slug.slice(index);
            slug = slug.slice(0, index);
          } else {
            path = '';
          }

          context$2$0.next = 10;
          return regeneratorRuntime.awrap(this._handleRequest(ctx, slug, path, next));

        case 10:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this._handleRequest = function callee$1$0(ctx, slug, path, next) {
    var registeredCollection;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          if (!(slug === '')) {
            context$2$0.next = 4;
            break;
          }

          context$2$0.next = 3;
          return regeneratorRuntime.awrap(this.handleGetRepositoryIdRequest(ctx));

        case 3:
          return context$2$0.abrupt('return');

        case 4:
          if (!(slug === 'ping')) {
            context$2$0.next = 8;
            break;
          }

          context$2$0.next = 7;
          return regeneratorRuntime.awrap(this.handlePingRequest(ctx));

        case 7:
          return context$2$0.abrupt('return');

        case 8:
          if (!(slug === 'authorizations')) {
            context$2$0.next = 12;
            break;
          }

          context$2$0.next = 11;
          return regeneratorRuntime.awrap(this.handleAuthorizationRequest(ctx, path, next));

        case 11:
          return context$2$0.abrupt('return');

        case 12:
          registeredCollection = this.registeredCollections[slug];

          if (!registeredCollection) {
            context$2$0.next = 17;
            break;
          }

          context$2$0.next = 16;
          return regeneratorRuntime.awrap(this.handleCollectionRequest(ctx, registeredCollection, path, next));

        case 16:
          return context$2$0.abrupt('return');

        case 17:
          context$2$0.next = 19;
          return regeneratorRuntime.awrap(co(next));

        case 19:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.readBody = function callee$1$0(ctx) {
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(co(regeneratorRuntime.mark(function callee$2$0() {
            return regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
              while (1) switch (context$3$0.prev = context$3$0.next) {
                case 0:
                  context$3$0.next = 2;
                  return parseBody.json(ctx, { limit: '8mb' });

                case 2:
                  return context$3$0.abrupt('return', context$3$0.sent);

                case 3:
                case 'end':
                  return context$3$0.stop();
              }
            }, callee$2$0, this);
          })));

        case 2:
          ctx.request.body = context$2$0.sent;

        case 3:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.authorizationUnserializer = function (obj) {
    // can be overridden
    return obj.query && obj.query.authorization;
  };

  this.verifyAuthorization = function callee$1$0(ctx, method, request) {
    var handler;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          handler = undefined;

          if (ctx.registeredCollection) {
            handler = ctx.registeredCollection.verifyAuthorizationHandler;
          }

          if (!handler) handler = this.verifyAuthorizationHandler;

          if (handler) {
            context$2$0.next = 5;
            break;
          }

          return context$2$0.abrupt('return', true);

        case 5:

          request = _.clone(request) || {};
          request.authorization = ctx.authorization;
          request.collection = ctx.collection;
          request.remoteCollection = ctx.remoteCollection;
          request.method = method;
          request.options = ctx.options;
          context$2$0.next = 13;
          return regeneratorRuntime.awrap(handler(request));

        case 13:
          return context$2$0.abrupt('return', context$2$0.sent);

        case 14:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.authorize = function callee$1$0(ctx, method, request) {
    var handler;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          handler = undefined;

          if (ctx.registeredCollection) {
            handler = ctx.registeredCollection.authorizeHandler;
          }

          if (!handler) handler = this.authorizeHandler;

          if (handler) {
            context$2$0.next = 5;
            break;
          }

          return context$2$0.abrupt('return', !!ctx.verifyAuthorizationResult);

        case 5:

          request = _.clone(request) || {};
          request.verifyAuthorizationResult = ctx.verifyAuthorizationResult;
          request.collection = ctx.collection;
          request.remoteCollection = ctx.remoteCollection;
          request.method = method;
          request.options = ctx.options;
          context$2$0.next = 13;
          return regeneratorRuntime.awrap(handler(request));

        case 13:
          return context$2$0.abrupt('return', context$2$0.sent);

        case 14:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.verifyAuthorizationAndAuthorize = function callee$1$0(ctx, method, request) {
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.verifyAuthorization(ctx, method, request));

        case 2:
          ctx.verifyAuthorizationResult = context$2$0.sent;
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(this.authorize(ctx, method, request));

        case 5:
          ctx.authorizeResult = context$2$0.sent;

          if (!ctx.authorizeResult) ctx['throw'](403, 'authorization failed');

        case 7:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  // === Repository requests ===

  this.handleGetRepositoryIdRequest = function callee$1$0(ctx) {
    var id;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'getRepositoryId'));

        case 2:
          context$2$0.next = 4;
          return regeneratorRuntime.awrap(this.repository.getRepositoryId());

        case 4:
          id = context$2$0.sent;

          ctx.body = { repositoryId: id };

        case 6:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handlePingRequest = function callee$1$0(ctx) {
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          ctx.body = 'pong';
          ctx.logLevel = 'silence';

        case 2:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  // === Authorization requests ===

  this.handleAuthorizationRequest = function callee$1$0(ctx, path, next) {
    var fragment, index, method;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          fragment = path;

          if (_.startsWith(fragment, '/')) fragment = fragment.slice(1);
          index = fragment.indexOf('/');

          if (index !== -1) {
            path = fragment.slice(index + 1);
            fragment = fragment.slice(0, index);
          } else {
            path = '';
          }
          method = ctx.method;

          if (!(method === 'POST' && !fragment && !path)) {
            context$2$0.next = 10;
            break;
          }

          context$2$0.next = 8;
          return regeneratorRuntime.awrap(this.handleSignInWithCredentialsRequest(ctx));

        case 8:
          context$2$0.next = 22;
          break;

        case 10:
          if (!(method === 'GET' && fragment && !path)) {
            context$2$0.next = 15;
            break;
          }

          context$2$0.next = 13;
          return regeneratorRuntime.awrap(this.handleSignInWithAuthorizationRequest(ctx, fragment));

        case 13:
          context$2$0.next = 22;
          break;

        case 15:
          if (!(method === 'DELETE' && fragment && !path)) {
            context$2$0.next = 20;
            break;
          }

          context$2$0.next = 18;
          return regeneratorRuntime.awrap(this.handleSignOutRequest(ctx, fragment));

        case 18:
          context$2$0.next = 22;
          break;

        case 20:
          context$2$0.next = 22;
          return regeneratorRuntime.awrap(co(next));

        case 22:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleSignInWithCredentialsRequest = function callee$1$0(ctx) {
    var handler, authorization;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 2:
          handler = this.signInWithCredentialsHandler;

          if (handler) {
            context$2$0.next = 5;
            break;
          }

          throw new Error('signInWithCredentialsHandler is undefined');

        case 5:
          context$2$0.next = 7;
          return regeneratorRuntime.awrap(handler(ctx.request.body));

        case 7:
          authorization = context$2$0.sent;

          if (!authorization) ctx['throw'](403, 'sign in with credentials failed');
          ctx.status = 201;
          ctx.type = 'application/json';
          ctx.body = JSON.stringify(authorization);

        case 12:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleSignInWithAuthorizationRequest = function callee$1$0(ctx, authorization) {
    var handler, isOkay;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          handler = this.signInWithAuthorizationHandler;

          if (handler) {
            context$2$0.next = 3;
            break;
          }

          throw new Error('signInWithAuthorizationHandler is undefined');

        case 3:
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(handler(authorization));

        case 5:
          isOkay = context$2$0.sent;

          if (!isOkay) ctx['throw'](403, 'sign in with authorization failed');
          ctx.status = 204;

        case 8:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleSignOutRequest = function callee$1$0(ctx, authorization) {
    var handler;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          handler = this.signOutHandler;

          if (handler) {
            context$2$0.next = 3;
            break;
          }

          throw new Error('signOutHandler is undefined');

        case 3:
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(handler(authorization));

        case 5:
          ctx.status = 204;

        case 6:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  // === Collection requests ===

  this.handleCollectionRequest = function callee$1$0(ctx, registeredCollection, path, next) {
    var name, fragment1, fragment2, index, camelCasedFragment1, camelCasedFragment2, method;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          name = registeredCollection.name;

          ctx.registeredCollection = registeredCollection;
          ctx.collection = this.repository.createCollection(name);
          ctx.remoteCollection = this.remoteRepository.createCollection(name);

          fragment1 = path;
          fragment2 = '';

          if (_.startsWith(fragment1, '/')) fragment1 = fragment1.slice(1);
          index = fragment1.indexOf('/');

          if (index !== -1) {
            fragment2 = fragment1.slice(index + 1);
            fragment1 = fragment1.slice(0, index);
          }
          camelCasedFragment1 = _.camelCase(fragment1);
          camelCasedFragment2 = _.camelCase(fragment2);
          method = ctx.method;

          if (!(method === 'POST' && fragment1 === 'get-items' && !fragment2)) {
            context$2$0.next = 17;
            break;
          }

          context$2$0.next = 15;
          return regeneratorRuntime.awrap(this.handleGetItemsRequest(ctx));

        case 15:
          context$2$0.next = 64;
          break;

        case 17:
          if (!(method === 'GET' && fragment1 === 'count' && !fragment2)) {
            context$2$0.next = 22;
            break;
          }

          context$2$0.next = 20;
          return regeneratorRuntime.awrap(this.handleCountItemsRequest(ctx));

        case 20:
          context$2$0.next = 64;
          break;

        case 22:
          if (!((method === 'GET' || method === 'POST') && ctx.registeredCollection.collectionMethods.hasOwnProperty(camelCasedFragment1) && !fragment2)) {
            context$2$0.next = 27;
            break;
          }

          context$2$0.next = 25;
          return regeneratorRuntime.awrap(this.handleCustomCollectionMethodRequest(ctx, camelCasedFragment1));

        case 25:
          context$2$0.next = 64;
          break;

        case 27:
          if (!((method === 'GET' || method === 'POST') && fragment1 && ctx.registeredCollection.itemMethods.hasOwnProperty(camelCasedFragment2))) {
            context$2$0.next = 32;
            break;
          }

          context$2$0.next = 30;
          return regeneratorRuntime.awrap(this.handleCustomItemMethodRequest(ctx, fragment1, camelCasedFragment2));

        case 30:
          context$2$0.next = 64;
          break;

        case 32:
          if (!(method === 'GET' && fragment1 && !fragment2)) {
            context$2$0.next = 37;
            break;
          }

          context$2$0.next = 35;
          return regeneratorRuntime.awrap(this.handleGetItemRequest(ctx, fragment1));

        case 35:
          context$2$0.next = 64;
          break;

        case 37:
          if (!(method === 'POST' && !fragment1 && !fragment2)) {
            context$2$0.next = 42;
            break;
          }

          context$2$0.next = 40;
          return regeneratorRuntime.awrap(this.handlePostItemRequest(ctx));

        case 40:
          context$2$0.next = 64;
          break;

        case 42:
          if (!(method === 'PUT' && fragment1 && !fragment2)) {
            context$2$0.next = 47;
            break;
          }

          context$2$0.next = 45;
          return regeneratorRuntime.awrap(this.handlePutItemRequest(ctx, fragment1));

        case 45:
          context$2$0.next = 64;
          break;

        case 47:
          if (!(method === 'DELETE' && fragment1 && !fragment2)) {
            context$2$0.next = 52;
            break;
          }

          context$2$0.next = 50;
          return regeneratorRuntime.awrap(this.handleDeleteItemRequest(ctx, fragment1));

        case 50:
          context$2$0.next = 64;
          break;

        case 52:
          if (!(method === 'GET' && !fragment1 && !fragment2)) {
            context$2$0.next = 57;
            break;
          }

          context$2$0.next = 55;
          return regeneratorRuntime.awrap(this.handleFindItemsRequest(ctx));

        case 55:
          context$2$0.next = 64;
          break;

        case 57:
          if (!(method === 'DELETE' && !fragment1 && !fragment2)) {
            context$2$0.next = 62;
            break;
          }

          context$2$0.next = 60;
          return regeneratorRuntime.awrap(this.handleFindAndDeleteItemsRequest(ctx));

        case 60:
          context$2$0.next = 64;
          break;

        case 62:
          context$2$0.next = 64;
          return regeneratorRuntime.awrap(co(next));

        case 64:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.emitEvent = function callee$1$0(ctx, event, request) {
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          if (!request) request = {};
          request.verifyAuthorizationResult = ctx.verifyAuthorizationResult;
          request.authorizeResult = ctx.authorizeResult;
          request.collection = ctx.collection;
          request.remoteCollection = ctx.remoteCollection;
          request.event = event;
          request.options = ctx.options;
          context$2$0.next = 9;
          return regeneratorRuntime.awrap(ctx.registeredCollection.emit(event, request));

        case 9:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this._getItem = function callee$1$0(ctx, id, errorIfMissing) {
    var item;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          if (errorIfMissing == null) errorIfMissing = ctx.options.errorIfMissing;
          if (errorIfMissing == null) errorIfMissing = true;
          if (!id) ctx['throw'](400, 'id required');
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(ctx.collection.getItem(id, { errorIfMissing: false }));

        case 5:
          item = context$2$0.sent;

          if (!item && errorIfMissing) ctx['throw'](404, 'item not found');
          return context$2$0.abrupt('return', item);

        case 8:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleGetItemRequest = function callee$1$0(ctx, id) {
    var item, remoteItem, className, remoteCollection;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this._getItem(ctx, id));

        case 2:
          item = context$2$0.sent;
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'getItem', { item: item }));

        case 5:
          remoteItem = undefined;

          if (item) {
            className = item['class'].name;
            remoteCollection = this.remoteRepository.createCollectionFromItemClassName(className);

            remoteItem = remoteCollection.unserializeItem(item);
          }
          context$2$0.next = 9;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didGetItem', { remoteItem: remoteItem, item: item }));

        case 9:
          if (remoteItem) {
            ctx.body = {
              'class': remoteItem['class'].name,
              value: remoteItem.serialize()
            };
          } else {
            ctx.status = 204;
          }

        case 10:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handlePostItemRequest = function callee$1$0(ctx) {
    var remoteItem;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 2:
          remoteItem = ctx.remoteCollection.unserializeItem(ctx.request.body);
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(ctx.collection.transaction((function callee$2$0() {
            var item;
            return regeneratorRuntime.async(function callee$2$0$(context$3$0) {
              while (1) switch (context$3$0.prev = context$3$0.next) {
                case 0:
                  item = ctx.collection.createItem(remoteItem);
                  context$3$0.next = 3;
                  return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'putItem', { remoteItem: remoteItem, item: item }));

                case 3:
                  context$3$0.next = 5;
                  return regeneratorRuntime.awrap(this.emitEvent(ctx, 'willPutItem', { remoteItem: remoteItem, item: item }));

                case 5:
                  context$3$0.next = 7;
                  return regeneratorRuntime.awrap(item.save(ctx.options));

                case 7:
                  remoteItem = ctx.remoteCollection.unserializeItem(item);
                  context$3$0.next = 10;
                  return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didPutItem', { remoteItem: remoteItem, item: item }));

                case 10:
                  ctx.status = 201;
                  ctx.body = {
                    'class': remoteItem['class'].name,
                    value: remoteItem.serialize()
                  };

                case 12:
                case 'end':
                  return context$3$0.stop();
              }
            }, null, this);
          }).bind(this)));

        case 5:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handlePutItemRequest = function callee$1$0(ctx, id) {
    var remoteItem;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 2:
          remoteItem = ctx.remoteCollection.unserializeItem(ctx.request.body);
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(ctx.collection.transaction((function callee$2$0() {
            var errorIfMissing, item;
            return regeneratorRuntime.async(function callee$2$0$(context$3$0) {
              while (1) switch (context$3$0.prev = context$3$0.next) {
                case 0:
                  errorIfMissing = ctx.options.createIfMissing ? false : undefined;
                  context$3$0.next = 3;
                  return regeneratorRuntime.awrap(this._getItem(ctx, id, errorIfMissing));

                case 3:
                  item = context$3$0.sent;
                  context$3$0.next = 6;
                  return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'putItem', { remoteItem: remoteItem, item: item }));

                case 6:
                  if (item) {
                    item.updateValue(remoteItem);
                  } else {
                    item = ctx.collection.unserializeItem(remoteItem);
                    item.isNew = false;
                  }
                  context$3$0.next = 9;
                  return regeneratorRuntime.awrap(this.emitEvent(ctx, 'willPutItem', { remoteItem: remoteItem, item: item }));

                case 9:
                  context$3$0.next = 11;
                  return regeneratorRuntime.awrap(item.save(ctx.options));

                case 11:
                  remoteItem = ctx.remoteCollection.createItem(item);
                  context$3$0.next = 14;
                  return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didPutItem', { remoteItem: remoteItem, item: item }));

                case 14:
                  ctx.body = {
                    'class': remoteItem['class'].name,
                    value: remoteItem.serialize()
                  };

                case 15:
                case 'end':
                  return context$3$0.stop();
              }
            }, null, this);
          }).bind(this)));

        case 5:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleDeleteItemRequest = function callee$1$0(ctx, id) {
    var hasBeenDeleted, item;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          hasBeenDeleted = false;
          context$2$0.next = 3;
          return regeneratorRuntime.awrap(this._getItem(ctx, id));

        case 3:
          item = context$2$0.sent;

          if (!item) {
            context$2$0.next = 15;
            break;
          }

          context$2$0.next = 7;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'deleteItem', { item: item }));

        case 7:
          context$2$0.next = 9;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'willDeleteItem', { item: item }));

        case 9:
          context$2$0.next = 11;
          return regeneratorRuntime.awrap(item['delete'](ctx.options));

        case 11:
          hasBeenDeleted = context$2$0.sent;

          if (!hasBeenDeleted) {
            context$2$0.next = 15;
            break;
          }

          context$2$0.next = 15;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didDeleteItem', { item: item }));

        case 15:
          ctx.type = 'application/json';
          ctx.body = hasBeenDeleted;

        case 17:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleGetItemsRequest = function callee$1$0(ctx) {
    var items, cache, remoteItems;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      var _this3 = this;

      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 2:
          context$2$0.next = 4;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'getItems'));

        case 4:
          context$2$0.next = 6;
          return regeneratorRuntime.awrap(ctx.collection.getItems(ctx.request.body, ctx.options));

        case 6:
          items = context$2$0.sent;
          cache = {};
          remoteItems = items.map(function (item) {
            var className = item['class'].name;
            var remoteCollection = _this3.remoteRepository.createCollectionFromItemClassName(className, cache);
            return remoteCollection.unserializeItem(item);
          });
          context$2$0.next = 11;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didGetItems', { remoteItems: remoteItems, items: items }));

        case 11:
          ctx.status = 201;
          ctx.body = remoteItems.map(function (remoteItem) {
            return {
              'class': remoteItem['class'].name,
              value: remoteItem.serialize()
            };
          });

        case 13:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleFindItemsRequest = function callee$1$0(ctx) {
    var items, cache, remoteItems;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      var _this4 = this;

      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'findItems'));

        case 2:
          context$2$0.next = 4;
          return regeneratorRuntime.awrap(ctx.collection.findItems(ctx.options));

        case 4:
          items = context$2$0.sent;
          cache = {};
          remoteItems = items.map(function (item) {
            var className = item['class'].name;
            var remoteCollection = _this4.remoteRepository.createCollectionFromItemClassName(className, cache);
            return remoteCollection.unserializeItem(item);
          });
          context$2$0.next = 9;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didFindItems', { remoteItems: remoteItems, items: items }));

        case 9:
          ctx.body = remoteItems.map(function (remoteItem) {
            return {
              'class': remoteItem['class'].name,
              value: remoteItem.serialize()
            };
          });

        case 10:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleCountItemsRequest = function callee$1$0(ctx) {
    var count;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'countItems'));

        case 2:
          context$2$0.next = 4;
          return regeneratorRuntime.awrap(ctx.collection.countItems(ctx.options));

        case 4:
          count = context$2$0.sent;
          context$2$0.next = 7;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didCountItems', { count: count }));

        case 7:
          ctx.type = 'application/json';
          ctx.body = count;

        case 9:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleFindAndDeleteItemsRequest = function callee$1$0(ctx) {
    var deletedItemsCount;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          context$2$0.next = 2;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, 'findAndDeleteItems'));

        case 2:
          context$2$0.next = 4;
          return regeneratorRuntime.awrap(ctx.collection.findAndDeleteItems(ctx.options));

        case 4:
          deletedItemsCount = context$2$0.sent;
          context$2$0.next = 7;
          return regeneratorRuntime.awrap(this.emitEvent(ctx, 'didFindAndDeleteItems'));

        case 7:
          ctx.type = 'application/json';
          ctx.body = deletedItemsCount;

        case 9:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleCustomCollectionMethodRequest = function callee$1$0(ctx, method) {
    var fn, result;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          if (!(ctx.method === 'POST')) {
            context$2$0.next = 3;
            break;
          }

          context$2$0.next = 3;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 3:
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, method));

        case 5:
          fn = ctx.registeredCollection.collectionMethods[method];

          if (fn === true) {
            fn = function callee$2$0(request) {
              return regeneratorRuntime.async(function callee$2$0$(context$3$0) {
                while (1) switch (context$3$0.prev = context$3$0.next) {
                  case 0:
                    context$3$0.next = 2;
                    return regeneratorRuntime.awrap(request.collection[method](request.options));

                  case 2:
                    context$3$0.t0 = context$3$0.sent;
                    return context$3$0.abrupt('return', {
                      body: context$3$0.t0
                    });

                  case 4:
                  case 'end':
                    return context$3$0.stop();
                }
              }, null, this);
            };
          }
          context$2$0.next = 9;
          return regeneratorRuntime.awrap(this._callCustomMethod(ctx, fn));

        case 9:
          result = context$2$0.sent;

          this._writeCustomMethodResult(ctx, result);

        case 11:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this.handleCustomItemMethodRequest = function callee$1$0(ctx, id, method) {
    var item, fn, result;
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          if (!(ctx.method === 'POST')) {
            context$2$0.next = 3;
            break;
          }

          context$2$0.next = 3;
          return regeneratorRuntime.awrap(this.readBody(ctx));

        case 3:
          context$2$0.next = 5;
          return regeneratorRuntime.awrap(this._getItem(ctx, id));

        case 5:
          item = context$2$0.sent;
          context$2$0.next = 8;
          return regeneratorRuntime.awrap(this.verifyAuthorizationAndAuthorize(ctx, method, { item: item }));

        case 8:
          fn = ctx.registeredCollection.itemMethods[method];

          if (fn === true) {
            fn = function callee$2$0(request) {
              return regeneratorRuntime.async(function callee$2$0$(context$3$0) {
                while (1) switch (context$3$0.prev = context$3$0.next) {
                  case 0:
                    context$3$0.next = 2;
                    return regeneratorRuntime.awrap(request.item[method](request.options));

                  case 2:
                    context$3$0.t0 = context$3$0.sent;
                    return context$3$0.abrupt('return', {
                      body: context$3$0.t0
                    });

                  case 4:
                  case 'end':
                    return context$3$0.stop();
                }
              }, null, this);
            };
          }
          context$2$0.next = 12;
          return regeneratorRuntime.awrap(this._callCustomMethod(ctx, fn, { item: item }));

        case 12:
          result = context$2$0.sent;

          this._writeCustomMethodResult(ctx, result);

        case 14:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this._callCustomMethod = function callee$1$0(ctx, fn) {
    var request = arguments.length <= 2 || arguments[2] === undefined ? {} : arguments[2];
    return regeneratorRuntime.async(function callee$1$0$(context$2$0) {
      while (1) switch (context$2$0.prev = context$2$0.next) {
        case 0:
          request.collection = ctx.collection;
          request.remoteCollection = ctx.remoteCollection;
          request.options = ctx.options;
          request.body = ctx.request.body;
          context$2$0.next = 6;
          return regeneratorRuntime.awrap(fn.call(this, request));

        case 6:
          return context$2$0.abrupt('return', context$2$0.sent);

        case 7:
        case 'end':
          return context$2$0.stop();
      }
    }, null, this);
  };

  this._writeCustomMethodResult = function (ctx, result) {
    if (result.body == null) {
      ctx.status = 204;
      return;
    }
    ctx.status = ctx.method === 'POST' ? 201 : 200;
    _.forOwn(result.headers, function (value, key) {
      key = _.kebabCase(key);
      ctx.response.set(key, value);
    });
    ctx.body = result.body;
  };

  this.getMiddleware = function (prefix) {
    if (!prefix) prefix = '';
    if (_.endsWith(prefix, '/')) prefix = prefix.slice(0, -1);
    var that = this;
    return regeneratorRuntime.mark(function callee$2$0(next) {
      var path;
      return regeneratorRuntime.wrap(function callee$2$0$(context$3$0) {
        while (1) switch (context$3$0.prev = context$3$0.next) {
          case 0:
            path = this.path;

            if (!prefix) {
              context$3$0.next = 7;
              break;
            }

            if (_.startsWith(path, prefix)) {
              context$3$0.next = 6;
              break;
            }

            context$3$0.next = 5;
            return next;

          case 5:
            return context$3$0.abrupt('return', context$3$0.sent);

          case 6:
            path = path.substr(prefix.length);

          case 7:
            context$3$0.next = 9;
            return that.handleRequest(this, path, next);

          case 9:
          case 'end':
            return context$3$0.stop();
        }
      }, callee$2$0, this);
    });
  };
});

module.exports = KindaRepositoryServer;

// monkey patched in kinda-repository-synchronizer/history-server.js